/*---------------------------------------------------------------------\
|                          ____ _   __ __ ___                          |
|                         |__  / \ / / . \ . \                         |
|                           / / \ V /|  _/  _/                         |
|                          / /__ | | | | | |                           |
|                         /_____||_| |_| |_|                           |
|                                                                      |
----------------------------------------------------------------------/
*
* This file contains private API, this might break at any time between releases.
* You have been warned!
*
*/
#ifndef ZYPPNG_MEDIA_HTTP_TRANSFERSETTINGS_H_INCLUDED
#define ZYPPNG_MEDIA_HTTP_TRANSFERSETTINGS_H_INCLUDED

#include <zypp-curl/TransferSettings>

namespace zyppng {
using TransferSettings = zypp::media::TransferSettings;
}

#endif
