/*---------------------------------------------------------------------\
|                          ____ _   __ __ ___                          |
|                         |__  / \ / / . \ . \                         |
|                           / / \ V /|  _/  _/                         |
|                          / /__ | | | | | |                           |
|                         /_____||_| |_| |_|                           |
|                                                                      |
\---------------------------------------------------------------------*/
// Automatically generated wrapper for zypp/Locale.h
// Source Namespace: zypp -> Target Namespace: zyppng
#ifndef LOCALE_H_ZYPPNG_WRAPPER_H
#define LOCALE_H_ZYPPNG_WRAPPER_H

#include <zypp/Locale.h>

namespace zyppng
{
    using ::zypp::Locale;
    using LocaleSet = std::unordered_set<Locale>;
} // namespace zyppng
#endif // LOCALE_H_ZYPPNG_WRAPPER_H
